# TODO coils, validate config
from pulr import config, register_puller, get_object_id
from pulr.converters import (parse_int, bit_to_data, int16_to_data,
                             int32_to_data, real32_to_data)

import pymodbus.client.sync
from functools import partial

client = None


def parse_offset(offset, addr):
    if isinstance(offset, int):
        bit = None
    else:
        if '/' in offset:
            offset, bit = offset.split('/')
            bit = int(bit)
        else:
            bit = None
        if offset.startswith('='):
            absolute = True
            offset = offset[1:]
        else:
            absolute = False
        offset = parse_int(offset)
        if absolute:
            offset -= addr
    if offset < 0:
        raise ValueError('offset is negative')
    if bit is not None and bit < 0:
        raise ValueError('bit is negative')
    return offset, bit


def process_data(fn, dtype):
    rr = fn()
    if rr.isError():
        raise Exception(f'Modbus error {fn}')
    return rr.bits if dtype == 'b' else rr.registers


def init(cfg_proto, cfg_pull, timeout=5):
    global client
    if cfg_proto['name'] in ['modbus/tcp', 'modbus/udp']:
        try:
            host, port = cfg_proto['source'].rsplit(':', 1)
        except:
            host = cfg_proto['source']
            port = 502
        if cfg_proto['name'] == 'modbus/tcp':
            client = pymodbus.client.sync.ModbusTcpClient(host, int(port))
        else:
            client = pymodbus.client.sync.ModbusUdpClient(host, int(port))
    else:
        raise ValueError(f'Unsupported protocol: {cfg_proto["name"]}')

    client.timeout = timeout

    for p in cfg_pull:
        try:
            unit = p['unit']
        except KeyError:
            unit = cfg_proto.get('default-unit', 1)
        u = parse_int(unit)
        reg = p['reg']
        addr = parse_int(reg[1:])
        if reg[0] == 'c':
            pfn = client.read_coils
        elif reg[0] == 'd':
            pfn = client.read_discrete_inputs
        elif reg[0] == 'h':
            pfn = client.read_holding_registers
        elif reg[0] == 'i':
            pfn = client.read_input_registers
        else:
            raise ValueError(f'Invalid register type: {reg[0]}')
        pmap = []
        for m in p.get('map'):
            offset = m['offset']
            digits = m.get('digits')
            o = get_object_id(m['id'])
            tp = m.get('type')
            if tp and '*' in tp:
                tp, multiplier = tp.split('*')
                tp = tp.strip()
                multiplier = float(multiplier.strip())
            else:
                multiplier = 1
            if reg[0] in ['h', 'i']:
                offset, bit = parse_offset(offset, addr)
                if bit is None:
                    if tp in ['real', 'real32']:
                        fn = partial(real32_to_data, o, offset, multiplier,
                                     digits)
                    elif not tp or tp in ['uint16', 'word']:
                        fn = partial(int16_to_data, o, offset, False,
                                     multiplier, digits)
                    elif tp in ['uint32', 'dword']:
                        fn = partial(int32_to_data, o, offset, False,
                                     multiplier, digits)
                    elif tp in ['sint16', 'int16']:
                        fn = partial(int16_to_data, o, offset, True, multiplier,
                                     digits)
                    elif tp in ['sint32', 'int32']:
                        fn = partial(int32_to_data, o, offset, True, multiplier,
                                     digits)
                    else:
                        raise ValueError(f'type unsupported: {tp}')
                else:
                    fn = partial(bit_to_data, o, offset, bit)
            pmap.append(fn)
        register_puller(
            partial(process_data,
                    partial(pfn, addr, count=p.get('count', 1), unit=u),
                    ('b' if reg[0] in ['c', 'd'] else 'w')), pmap)
    client.connect()

def shutdown():
    client.close()



# TODO (SNMP trap, validate output config)

from functools import partial
from types import SimpleNamespace
from .beacons import beacon_eva_datapuller

oprint = partial(print, flush=True)

output_params = {}

_d = SimpleNamespace()


def output_devnull(*args, **kwargs):
    pass


def output_stdout(o, value):
    oprint(f'{o} {value}')


def output_stdout_ndjson(o, value):
    q = '' if isinstance(value, int) or isinstance(value, float) else '"'
    oprint(f'{{"id":"{o}","v":{q}{value}{q}}}')


def output_webhook(o, value):
    try:
        s = _d.webs
    except AttributeError:
        import requests
        _d.webs = requests.Session()
        s = _d.webs
    from pulr import config
    r = s.post(output_params['url'],
               json={
                   'id': o,
                   'v': value
               },
               headers=output_params.get('headers'),
               timeout=output_params['timeout'])
    if not r.ok:
        raise RuntimeError(
            f'webhook {output_params["url"]} response {r.status_code}')


def output_eva_datapuller(o, value):
    s = f'{o[0]} u '
    oprint(s + (str(value) if o[1] == 's' else f'None {value}'))


def send_beacon_eva_datapuller():
    oprint()


OUTPUT_METHODS = {
    'stdout': {
        'output': output_stdout
    },
    'stdout/ndjson': {
        'output': output_stdout_ndjson
    },
    'stdout/eva-datapuller': {
        'output': output_eva_datapuller,
        'beacon': beacon_eva_datapuller
    },
    'webhook': {
        'output': output_webhook
    }
}



from pulr import set_data

import struct

def parse_int(i):
    if isinstance(i, int):
        return i
    elif 'x' in i:
        return int(i, 16)
    else:
        return int(i)

# common data postprocessors

def int16_to_data(o, offset, signed, multiplier, digits, data_in):
    value = data_in[offset]
    if signed and value > 32767:
        value -= 65536
    value *= multiplier
    if digits is not None:
        value = round(value, digits)
    set_data(o, value)


def int32_to_data(o, offset, signed, multiplier, digits, data_in):
    value = data_in[offset] * 65536 + data_in[offset + 1]
    if signed and value > 2147483647:
        value -= 4294967296
    value *= multiplier
    if digits is not None:
        value = round(value, digits)
    set_data(o, value)


def real32_to_data(o, offset, multiplier, digits, data_in):
    value = struct.unpack(
        'f',
        struct.pack('H', data_in[offset]) +
        struct.pack('H', data_in[offset + 1]))[0] * multiplier
    if digits is not None:
        value = round(value, digits)
    set_data(o, value)


def bit_to_data(o, offset, bit, data_in):
    x = data_in[offset]
    set_data(o, (x >> bit) & 1)




__version__ = '0.0.1'

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='pulr',
    version=__version__,
    author='Altertech',
    author_email='div@altertech.com',
    description='Industrial protocols data puller',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/alttch/pulr',
    packages=setuptools.find_packages(),
    license='Apache License 2.0',
    install_requires=['pyyaml', 'jsonschema'],
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Communications',
        'Topic :: System :: Networking :: Monitoring :: Hardware Watchdog',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces'))



__author__ = 'Altertech, https://www.altertech.com/'
__copyright__ = 'Copyright (C) 2020 Altertech'
__license__ = 'Apache License 2.0'

__version__ = '0.0.1'

# TODO: validate config

import sys
import argparse
import importlib
import threading
import jsonschema

import yaml

from time import perf_counter, sleep
from .outputs import OUTPUT_METHODS, output_params, oprint

data = {}

pullers = []
processor_maps = []

DEFAULT_TIMEOUT = 5
DEFAULT_FREQUENCY = 1
DEFAULT_BEACON_FREQUENCY = 2

config = {
    'timeout': DEFAULT_TIMEOUT,
    'freq': DEFAULT_FREQUENCY,
    'beacon': DEFAULT_BEACON_FREQUENCY,
    'output': {
        'type': 'stdout'
    }
}

output = None


def register_puller(fn, pmap=[]):
    pullers.append((fn, pmap))


def get_object_id(i):
    if config['output']['type'] == 'stdout/eva-datapuller':
        o = i.rsplit('.', 1)
        if o[1] == 'status':
            o[1] = 's'
        elif o[1] == 'value':
            o[1] = 'v'
        o = tuple(o)
    else:
        o = i
    return o


def set_data(o, value):
    current = data.get(o)
    if current != value:
        data[o] = value
        output(o, value)


def _t_beacon(fn, interval):
    try:
        next_beacon = perf_counter() + interval
        while True:
            ts = next_beacon - perf_counter()
            if ts > 0:
                sleep(ts)
            fn()
            next_beacon += interval
    except:
        import traceback
        oprint('beacon error', file=sys.stderr)
        oprint(traceback.format_exc())


def do(loop=False):

    interval = config['interval']

    next_iter = perf_counter() + interval

    def pull_and_process():
        nonlocal next_iter
        for plr, prc_map in pullers:
            data = plr()
            for fn in prc_map:
                fn(data)
        ts = next_iter - perf_counter()
        if loop:
            if ts > 0:
                sleep(ts)
            else:
                oprint('WARNING: main loop timeout', file=sys.stderr)
        next_iter += interval

    if loop:
        while True:
            pull_and_process()
    else:
        pull_and_process()


def main():
    global output
    ap = argparse.ArgumentParser()
    ap.add_argument('-F',
                    '--config',
                    help='Configuration file',
                    metavar='CONFIG',
                    required=True)
    ap.add_argument('-L',
                    '--loop',
                    help='Loop (production)',
                    action='store_true')
    a = ap.parse_args()

    with open(a.config) as fh:
        config.update(yaml.safe_load(fh))

    config['interval'] = 1 / config['freq']

    try:
        om = OUTPUT_METHODS[config['output']['type']]
    except KeyError:
        raise Exception('Unsupported output type or output type not specified')
    output = om['output']
    output_params.update(config['output'])
    send_beacon = om.get('beacon')

    if 'timeout' not in output_params:
        output_params['timeout'] = config['timeout']

    proto = config['proto']['name']

    if '/' in proto:
        proto = proto.split('/', 1)[0]

    lib = importlib.import_module(f'pulr.proto.{proto}')

    lib.init(config['proto'], config.get('pull', []), timeout=config['timeout'])

    if a.loop and send_beacon:
        threading.Thread(target=_t_beacon,
                         name='beacon',
                         args=(send_beacon, config['beacon']),
                         daemon=True).start()

    try:
        do(loop=a.loop)
    finally:
        lib.shutdown()
